
#include "instr64.h"
#include "mem.h"


#define BIAS	(2048-1)
#define FRSIZE	176

GLOBL	ofwentry(SB), $4

#define CRASH		NOSCHED; WORD $1; SCHED

/*
 * Entry point from openfirmware. 
 * We're called with the client interface handle in %o4
 * and our stack pointer in %o6.  
 * We start at 0x4000 virtual which is also 0x4000 physical.
 * Since we're 64-bits, %o6 is offset by BIAS.
 */
TEXT start(SB),$-4
	ADD		$BIAS, R14, R1

	/* 0..4M at KZERO in itlb and dtlb */
	MOVW	$MMU_TAGACCESS, R8
	MOVW	$KZERO, R9
	STXAr(9, 8, 0, ASI_MMU_I)
	STXAr(9, 8, 0, ASI_MMU_D)
	MEMBAR(MBSYNC)
	MOVW	$0xe0000000, R8
	SLLXi(8, 8, 32)
	OR		$(PTELOCK|PTEKERNEL|PTEWRITE|PTEGLOB | 0x30), R8
	MOVW	$(63<<3), R9
	STXAr(8, 9, 0, ASI_MMU_IDATA)
	STXAr(8, 9, 0, ASI_MMU_DDATA)
	MEMBAR(MBSYNC)

	MOVW	$startvirt(SB), R7
	JMPL		(R7)

TEXT startvirt(SB), $-4
	/* setup R1, R2 and ofwentry */
	MOVW	$setSB(SB), R2
	MOVW	R12, ofwentry(SB)

	WRPRi(RPIL, 0, 15)
	WRPRi(RPSTATE, 0, PSPEF|PSAM|PSPRIV|PSIE)
	WRPRi(RCWP, 0, 0)
	WRSTi(RASI, 0, 0);

	/* XXX setup FSR */
	FMOVD	$0.5, F26			/* 0.5 -> F26 */
	FSUBD	F26, F26, F24		/* 0.0 -> F24 */
	FADDD	F26, F26, F28		/* 1.0 -> F28 */
	FADDD	F28, F28, F30		/* 2.0 -> F30 */
	FMOVD	F24, F0
	FMOVD	F24, F2
	FMOVD	F24, F4
	FMOVD	F24, F6
	FMOVD	F24, F8
	FMOVD	F24, F10
	FMOVD	F24, F12
	FMOVD	F24, F14
	FMOVD	F24, F16
	FMOVD	F24, F18
	FMOVD	F24, F20
	FMOVD	F24, F22

	JMPL		main(SB)
	CRASH

/* DEBUG STUFF
TEXT getitlb(SB), $0
	MOVW	8(R1), R8
	LDXAr(9, 7, 0, ASI_MMU_ITAG)
	STXi	(9, 8, 0)
	LDXAr(9, 7, 0, ASI_MMU_IDATA)
	STXi	(9, 8, 8)
	RETURN

TEXT getdtlb(SB), $0
	MOVW	8(R1), R8
	LDXAr(9, 7, 0, ASI_MMU_DTAG)
	STXi	(9, 8, 0)
	LDXAr(9, 7, 0, ASI_MMU_DDATA)
	STXi	(9, 8, 8)
	RETURN
*/

TEXT getasi4(SB), $0
	WRSTi(RASI, 7, 0)
	MOVW	8(R1), R8
	LDWAi(7, 8, 0)
	RETURN

TEXT getasi8(SB), $0
	MOVW	8(R1), R8
	WRSTi(RASI, 8, 0)
	MOVW	12(R1), R8
	LDXAi(9, 8, 0)
	STXi(9, 7, 0)
	RETURN

TEXT putasi4(SB), $0
	WRSTi(RASI, 7, 0)
	MOVW	8(R1), R8
	MOVW	12(R1), R9
	STWAi(9, 8, 0)
	RETURN

TEXT putasi8(SB), $0
	WRSTi(RASI, 7, 0)
	MOVW	8(R1), R8
	MOVW	12(R1), R9
	MOVW	16(R1), R10
	SLLXi(8, 8, 32)
	OR		R10, R9
	STXAi(9, 8, 0)
	RETURN

TEXT icflush(SB), $0
	FLUSHr(7, 0)
	RETURN

TEXT puttba(SB), $0
	WRSTi(RTBA, 7, 0)
	RETURN

/* port: test-and-set */
TEXT tas(SB), $0
	TAS (R7), R7
	RETURN

/* port: atomic increment */
TEXT _xinc(SB), $0
  xincagain:
	MOVW	(R7), R8
	ADD		$1, R8, R9
	CASAr(9, 7, 8, ASI_PRIMARY)
	BNE		xincagain
	RETURN

/* port: atomic decrement */
TEXT _xdec(SB), $0
 xdecagain:
	MOVW	(R7), R8
	ADD		$-1, R8, R9
	CASAr(9, 7, 8, ASI_PRIMARY)
	BNE		xdecagain
	RETURN

/* 
 * port: note: order here is important.  port depends on spllo and spldone addresses.
 */
TEXT spllo(SB), $0
	RDPR(7, RPIL)
	WRPRi(RPIL, 0, 0)
	RETURN

/* port */
TEXT splhi(SB), $0
	MOVW	R15, 4(R(MACH))	/* XXX:offset.  m->splpc = pc */
	RDPR(7, RPIL)
	WRPRi(RPIL, 0, 15)
	RETURN

/* port */
TEXT splx(SB), $0
	MOVW	R15, 4(R(MACH))	/* XXX:offset.  m->splpc = pc */
	WRPRi(RPIL, 7, 0)
	RETURN

/* port: end of spl stuff */
TEXT spldone(SB), $0
	RETURN

/* port: are we spllo? */
TEXT islo(SB), $0
	RDPR(8, RPIL)
	MOVRi(7, 8, 0, RCOND_Z)
	RETURN

/* call openfirmware */
TEXT ofcall(SB), $0
	/* store caller-saveds in local regs */
	MOVW	R1, R17
	MOVW	R2, R18
	MOVW	R3, R19
	MOVW	R4, R20
	MOVW	R5, R21
	MOVW	R6, R22

	/* get sparc64 stack pointer, and perform call with our arg */
	MOVW	R7, R8
	MOVW	ofwentry(SB), R16
	ANDN	$0xf, R1
	SUB		$(BIAS + FRSIZE), R1, R14
	JMPL		(R16)

	/* restore our registers (including SP, no need for fixup) and return */
	MOVW	R17, R1
	MOVW	R18, R2
	MOVW	R19, R3
	MOVW	R20, R4
	MOVW	R21, R5
	MOVW	R22, R6
	MOVW	R8, R7
	RETURN

/* port: given addr of first arg, fetch the return address */
TEXT getcallerpc(SB), $0
	MOVW	-4(R7), R7
	RETURN

/* port: ensure memory ordering */
TEXT coherence(SB), $0
	MEMBAR(MBSYNC)
	RETURN

/* port */
TEXT setlabel(SB), $0
	MOVW	R1, (R7)
	MOVW	R15, 4(R7)
	MOVW	$0, R7
	RETURN

/* port */
TEXT gotolabel(SB), $0
	MOVW	(R7), R1
	MOVW	4(R7), R15
	MOVW	$1, R7
	RETURN

